<?php
/*
S-GAE (Simple Graphical Accounting Engine) is a web application developed at /rdlab to 
graphically display the accounting information generated by our HPC queue manager.

Copyright (C) +2020  rdlab@cs.upc.edu - https://rdlab.cs.upc.edu

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

namespace sgae3\Controller;

use Exception;

class RouteController
{

    private $get_routes;
    private $post_routes;
    private $all_routes;
    private $some_routes;

    public function __construct()
    {
        $this->get_routes = array();
        $this->post_routes = array();
        $this->all_routes = array();
        $this->some_routes = array();
    }

    //Add callback for regular GET or POST routes
    public function addRoute(string $type, string $route, string $class, string $function)
    {
        if ($type == "GET") {
            $this->addGET($route, $class, $function);
        } elseif ($type == "POST") {
            $this->addPOST($route, $class, $function);
        } else {
            throw new Exception("Route method not supported");
        }
    }

    //Add callback for all routes
    public function addAll(string $class, string $function)
    {
        $this->all_routes[] = array($class, $function);
    }

    //Add callback for some routes (e.g. /admin)
    public function addSome(string $route, string $class, string $function)
    {
        $this->some_routes[$route] = array($class, $function);
    }

    private function addGET(string $route, string $class, string $function)
    {
        $this->get_routes[$route] = array($class, $function);
    }

    private function addPOST(string $route, string $class, string $function)
    {
        $this->post_routes[$route] = array($class, $function);
    }

    public function parseRequest(): array
    {
        // Get route
        $base = dirname($_SERVER['PHP_SELF']);
        $request_uri = $_SERVER['REQUEST_URI'];
        if (ltrim($base, '/')) {
            // Update request when we have a subdirectory
            $request_uri = substr($request_uri, strlen($base));
        }

        // Get method
        $request_method = $_SERVER['REQUEST_METHOD'];

        return [$request_uri, $request_method];
    }

    public function dispatch()
    {
        list($request_uri, $request_method) = $this->parseRequest();

        // These apply to all
        foreach ($this->all_routes as $all) {
            list($controller_name, $method) = $all;
            $controller = __NAMESPACE__ . "\\" . $controller_name;
            $c = new $controller;
            $c->$method();
        }

        // These apply only to matching (e.g /admin)
        foreach ($this->some_routes as $route => $action) {
            if (strpos($request_uri, $route)) {
                list($controller_name, $method) = $action;
                $controller = __NAMESPACE__ . "\\" . $controller_name;
                $c = new $controller;
                $c->$method();
            }
        }

        if ($request_method == "GET") {
            $this->dispatchGET($request_uri);
        } elseif ($request_method == "POST") {
            $this->dispatchPOST($request_uri);
        } else {
            trigger_error("Route method $request_method not supported");
            $this->return405();
        }
    }

    private function dispatchGet(string $request_uri)
    {
        require $GLOBALS['app_path'] . '/views/layout/page-top.phtml';

        if (array_key_exists($request_uri, $this->get_routes)) {
            list($controller_name, $method) = $this->get_routes[$request_uri];
            $controller = __NAMESPACE__ . "\\" . $controller_name;
            $c = new $controller;
            $c->$method();
        } else {
            $this->return404();
        }

        require $GLOBALS['app_path'] . '/views/layout/page-bottom.phtml';
    }

    private function dispatchPost(string $request_uri)
    {
        if (array_key_exists($request_uri, $this->post_routes)) {
            list($controller_name, $method) = $this->post_routes[$request_uri];
            $controller = __NAMESPACE__ . "\\" . $controller_name;
            $c = new $controller;
            $c->$method();
        } else {
            $this->return404();
        }
    }

    /*private function return401()
    {
        http_response_code(401);
        require $GLOBALS['app_path'] . '/views/401.phtml';
        die();
    }*/

    /*private function return403()
    {
        http_response_code(403);
        require $GLOBALS['app_path'] . '/views/403.phtml';
        die();
    }*/

    public function return404()
    {
        http_response_code(404);
        require $GLOBALS['app_path'] . '/views/404.phtml';
        die();
    }

    private function return405()
    {
        http_response_code(405);
        require $GLOBALS['app_path'] . '/views/405.phtml';
        die();
    }

    /**
     * Add-on for modules
     * As modules classes are outside composer's autoloader,
     * instead of class name the route gets a class instance.
     */
    //Add callback for regular GET or POST routes
    public function addModuleRoute(string $type, string $route, object $class, string $function)
    {
        if ($type == "GET") {
            $this->addModuleGET($route, $class, $function);
        } elseif ($type == "POST") {
            $this->addModulePOST($route, $class, $function);
        } else {
            throw new Exception("Route method not supported");
        }
    }

    private function addModuleGET(string $route, object $class, string $function)
    {
        $this->get_routes[$route] = array($class, $function);
    }

    private function addModulePOST(string $route, object $class, string $function)
    {
        $this->post_routes[$route] = array($class, $function);
    }

    public function moduleDispatch()
    {
        list($request_uri, $request_method) = $this->parseRequest();

        if ($request_method == "GET") {
            $this->moduleDispatchGET($request_uri);
        } elseif ($request_method == "POST") {
            $this->moduleDispatchPOST($request_uri);
        } else {
            trigger_error("Route method $request_method not supported");
            $this->return405();
        }
    }

    private function moduleDispatchGet(string $request_uri)
    {
        require $GLOBALS['app_path'] . '/views/layout/page-top.phtml';

        if (array_key_exists($request_uri, $this->get_routes)) {
            list($controller, $method) = $this->get_routes[$request_uri];
            $controller->$method();
        } else {
            $this->return404();
        }

        require $GLOBALS['app_path'] . '/views/layout/page-bottom.phtml';
    }

    private function moduleDispatchPost(string $request_uri)
    {
        if (array_key_exists($request_uri, $this->post_routes)) {
            list($controller, $method) = $this->post_routes[$request_uri];
            $controller->$method();
        } else {
            $this->return404();
        }
    }
}
