<?php
/*
S-GAE (Simple Graphical Accounting Engine) is a web application developed at /rdlab to 
graphically display the accounting information generated by our HPC queue manager.

Copyright (C) +2020  rdlab@cs.upc.edu - https://rdlab.cs.upc.edu

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

namespace sgae3\Helper;

use Exception;

class Sacct
{
    var $fields;
    var $finished_states;

    function __construct()
    {
        // https://slurm.schedmd.com/sacct.html
        $this->fields = array(
            "Partition",
            "User",
            "JobID",
            "JobName",
            "Submit",
            "Start",
            "End",
            "ReqTRES",
            "NodeList",
            "NCPUS",
            "CPUTimeRAW",
            "ElapsedRaw",
            "MaxDiskRead",
            "MaxDiskWrite",
            "MaxVMSize",
            "State",
            "ExitCode"
        );

        // https://slurm.schedmd.com/sacct.html#lbAGF
        $this->finished_states = array(
            "BOOT_FAIL",
            "CANCELLED",
            "COMPLETED",
            "DEADLINE",
            "FAILED",
            "NODE_FAIL",
            "OUT_OF_MEMORY",
            "PREEMPTED",
            "REVOKED",
            "TIMEOUT"
        );
    }

    /**
     * Main class entrypoint.
     * @param string $queue: queue name
     * @param string $from: format: 2020-02-19T15:37:14
     * @param string $to: format: 2020-02-19T15:42:21
     */
    public function import(string $queue, string $from, string $to)
    {
        //Create file to work in
        $fd = fopen($GLOBALS['output_file'], "w");
        fclose($fd);

        // BUG: sacct / slurmdbd may crash with large ammounts of jobs
        // HACK: do smaller daily queries

        list($from_date_part, $from_hour_part) =  explode("T", $from);
        list($to_date_part, $to_hour_part) =  explode("T", $to);
        list($start_year, $start_month, $start_day) = explode("-", $from_date_part);
        list($end_year, $end_month, $end_day) = explode("-", $to_date_part);

        for ($year = $start_year; $year <= $end_year; $year++) {

            list($month_begin, $month_end) = [1, 12];
            if ($year == $start_year) {
                $month_begin = $start_month;
            }
            if ($year == $end_year) {
                $month_end = $end_month;
            }

            for ($month = $month_begin; $month <= $month_end; $month++) {
                $day_begin = 1;
                $date = new \DateTime("last day of $year-$month");
                $day_end = $date->format('d');
                if ($year == $start_year && $month == $start_month) {
                    $day_begin = $start_day;
                }
                if ($year == $end_year && $month == $end_month) {
                    $day_end = $end_day;
                }

                if (strlen($month) < 2) $month = "0$month";

                for ($day = $day_begin; $day <= $day_end; $day++) {
                    if ($year == $start_year && $month == $start_month && $day == $start_day) {
                        $hour_begin = $from_hour_part;
                        $hour_end = "23:59:59";
                    } else if ($year == $end_year && $month == $end_month && $day == $end_day) {
                        $hour_begin = "00:00:00";
                        $hour_end = $to_hour_part;
                    } else {
                        $hour_begin = "00:00:00";
                        $hour_end = "23:59:59";
                    }

                    $this->sacct($queue, $day, "$year-$month-${day}T$hour_begin", "$year-$month-${day}T$hour_end");
                }
            }
        }

        //Delete output file
        unlink($GLOBALS['output_file']);
    }

    /**
     * Invokes Slurm's sacct; contanins HACK due to sacct inestability and bugs.
     *
     * @param string $queue: queue name
     * @param int $day: the day we are processing. As a param for optimization.
     * @param string $from: format: 2020-02-19T15:37:14
     * @param string $to: format: 2020-02-19T15:42:21
     */
    public function sacct(string $queue, int $day, string $from, string $to)
    {

        $sacct = $GLOBALS['sacct_path'];
        $fields = implode(",", $this->fields);
        $output_file = $GLOBALS['output_file'];

        // BUG: sacct's flag '-s' does not work on states different than COMPLETED
        // HACK: get 'State' and filter with egrep
        $states_egrep = implode("|", $this->finished_states);

        // First, get sacct output and leave it in a file
        // -r: partition; -S: Start; -E: End; -o: field_list; -a: all_users; -n: no_headers
        // -P: do not use last "|" (ignored); --units=K: show values en KB. 
        $command = "$sacct -r '$queue' -S '$from' -E '$to' -o $fields -a -P -n --units=K > $output_file";
        $output = array();
        $return_var = null;
        exec($command, $output, $return_var);
        if ($return_var != 0) {
            trigger_error("ERROR: command '$command' ended with status $return_var and output: " . implode(",", $output), E_USER_ERROR);
            throw new Exception("Cron exception: see log for details");
        }

        // Grep and count output lines. Otherwise, grep returns status = 1, so skip.
        $command = "cat $output_file | egrep \"$states_egrep\" | wc -l";
        $output = array();
        $return_var = null;
        exec($command, $output, $return_var);
        if ($return_var != 0) {
            trigger_error("ERROR: command '$command' ended with status $return_var and output: " . implode(",", $output), E_USER_ERROR);
            throw new Exception("Cron exception: see log for details");
        }

        // Do grep
        if ($output[0] > 0) {
            $command = "cat $output_file | egrep \"$states_egrep\"";
            $output = array();
            $return_var = null;
            exec($command, $output, $return_var);
            if ($return_var != 0) {
                trigger_error("ERROR: command '$command' ended with status $return_var and output: " . implode(",", $output), E_USER_ERROR);
                throw new Exception("Cron exception: see log for details");
            }

            // We have some output: do something with it
            $this->get_jobs_data($output, $day);
        } else {
            //echo "No jobs on queue '$queue' from '$from' to '$to'" . PHP_EOL;
        }
    }



    /**
     * Get accounting information from import() output.
     * Output contains data form multiple jobs: group it, then process.
     * @param array $output: sacct data.
     * @param int $day: the day we are processing. As a param for optimization.
     */
    private function get_jobs_data(array $output, int $day)
    {

        // Get a header => value array
        $named = $this->transform_output_to_named_array($this->fields, $output);

        $this_job = array();
        $num_elems = count($named);
        for ($i = 0; $i < count($named); $i++) {
            // Keep job data, join with next lines if they belong to same job
            $this_job[] = $named[$i];
            if ($i + 1 == $num_elems) {
                //Last element: process
                $this->process($this_job, $day);
                // Clean
                $this_job = array();
            } else {
                $this_job_base_jobname = explode(".", $named[$i]['JobID'])[0];
                $next_job_base_jobname = explode(".", $named[$i + 1]['JobID'])[0];

                // Next element is another job
                if ($this_job_base_jobname != $next_job_base_jobname) {
                    $this->process($this_job, $day);
                    // Clean
                    $this_job = array();
                }
            }
        }
    }
    /**
     * @param array $job_data: (multiline) data from a job
     * @param int $day: the day we are processing. As a param for optimization.
     */
    private function process(array $job_data, int $day)
    {

        $num_elems = count($job_data);
        if ($num_elems == 0) {
            // Job has no accounting data
            return;
        }

        // Avoid jobs started yesterday and ending today
        // They should have been parsed on "yesterday" iteration
        $line = $job_data[0];
        $start_day = date("d", $this->make_timestamp($line["Start"]));
        if ($start_day != $day) {
            return;
        }

        if ($num_elems == 1) {
            //One-liner (heterogeneus jobs are one-liners too)
            $this->parse_one_row_job($job_data);
            return;
        }

        $has_batch = false;
        $has_steps = false;

        // Multi-row job: read each line's JobID and determine what to do
        for ($i = 0; $i < count($job_data); $i++) {

            $line = $job_data[$i];
            $job_id = $line["JobID"];

            /* Rolled-up array jobs: 3 cases detected   
            * - range: 601662_[315-20000]) 
            * - one-element range: 2607476_[18]
            * - list: 4363_[0-1,3]
            * must be rolled-down:
            */
            if (preg_match("/([0-9]+)_\[(.*)\]/", $job_id, $matches)) {
                array_shift($matches);
                list($job_id_base, $job_id_range) = $matches;
                // Unroll grouped id's and process this new list
                $unrolled_job_ids = array();
                $this->unroll($job_id_base, $job_id_range, $unrolled_job_ids);
                $new_job_data = $line;
                foreach ($unrolled_job_ids as $unrolled_job_id) {
                    $new_job_data["JobID"] = $unrolled_job_id;
                    $this->parse_one_row_job($new_job_data);
                }
                unset($new_job_data);
                // We are done
                return;
            }

            // Non rolled-up array jobs are in the form: ArrayJobID_ArrayTaskID, followed by ArrayJobID_ArrayTaskID.batch
            if (strpos($job_id, "_")) {
                $new_job_data = array();
                $new_job_data[] = $line;
                //Get also next element
                $new_job_data[] = $job_data[++$i];
                $this->parse_job_with_dot_bash($new_job_data);
                // We are done
                return;
            }

            // Regular jobs (having .batch) or jobs with steps (JobID[.JobStep])
            else if (strpos($job_id, ".batch")) {
                $has_batch = true;
            } else if (preg_match('/\.\d/', $job_id)) {
                $has_steps = true;
            }
        }

        if ($has_batch) {
            $this->parse_job_with_dot_bash($job_data);
        } else if ($has_steps) {
            $this->parse_steps_only_job($job_data);
        } else {
            // Inform but don't abort
            trigger_error("Unexpected kind of job: " . $job_data[0]["JobID"], E_USER_ERROR);
        }
    }

    private function unroll(string $job_id_base, string $job_id_range, array &$unrolled_job_ids)
    {
        //Case 4363_[0-1,3]
        if (strpos($job_id_range, ",")) {
            foreach (explode(",", $job_id_range) as $section) {
                $this->unroll($job_id_base, $section, $unrolled_job_ids);
            }
        }
        //Case 601662_[315-320]
        else if (strpos($job_id_range, "-")) {
            list($first, $last) = explode("-", $job_id_range);
            for ($i = $first; $i <= $last; $i++) {
                $unrolled_job_ids[] = $job_id_base . "_$i";
            }
        }
        //Case 2607476_[18]
        else {
            $unrolled_job_ids[] = "${job_id_base}_${job_id_range}";
        }
    }

    /**
     * For each sacct output line from a job, create an array with $headers as keys and $output as values
     *
     * @param array[] $headers: first row of sacct output, name of each field
     * @param array[] $output: job data from sacct output, actual job stats information
     * @return array[] $named: the array with $headers as keys and $output as values
     */
    private function transform_output_to_named_array($headers, $output): array
    {

        $num_headers = count($headers);
        $named = array();
        foreach ($output as $i => $line) {
            $named[$i] = array();
            $data = explode("|", $line);

            // HACK: -P Sets an extra "|" delimiter in case of more than one result
            $num_cols = count($data);
            if ($num_cols == ($num_headers + 1)) {
                array_pop($data);
            }

            foreach ($data as $index => $field) {
                $named[$i][$headers[$index]] = trim($field);
            }
        }
        return $named;
    }

    /**
     * This job prints only 1 line on sacct:
     * JobID State ExitCode
     * ------------ ----------- --------
     * %job_id% COMPLETED 0:0
     *
     * @param array[] $job: output from transform_output_to_named_array
     */
    private function parse_one_row_job($job)
    {

        $data = array();
        $line = $job[0];
        $data["Partition"] = $line["Partition"];
        $data["User"] = $line["User"];
        $data["JobID"] = $line["JobID"];
        $data["JobName"] = $line["JobName"];
        $data["Submit"] = $this->make_timestamp($line["Submit"]);
        $data["ReqTRES"] = $this->parse_reqTRES($line["ReqTRES"]);
        $data["NodeList"] = $line["NodeList"];
        $data["NCPUS"] = $line["NCPUS"];
        $data["CPUTimeRAW"] = $line["CPUTimeRAW"];
        $data["ElapsedRaw"] = $line["ElapsedRaw"];
        $data["ExitCode"] = $line["ExitCode"];
        $data["Start"] = $this->make_timestamp($line["Start"]);
        $data["End"] = $this->make_timestamp($line["End"]);
        $data["MaxDiskRead"] = $this->parse_unit_in_KB($line["MaxDiskRead"]);
        $data["MaxDiskWrite"] = $this->parse_unit_in_KB($line["MaxDiskWrite"]);
        $data["MaxVMSize"] = $line["MaxVMSize"];
        $data["State"] = $line["State"];

        // If no MaxVM provided, use reserved
        if (trim($data["MaxVMSize"]) == "") {
            $data["MaxVMSize"] = $data["ReqTRES"];
        } else {
            $data["MaxVMSize"] = $this->parse_unit_in_KB($data["MaxVMSize"]);
        }

        unset($job);
        unset($line);
        $this->save($data);
    }

    /**
     * This job prints >1 lines on sacct.
     * It has one of these patterns:
     *
     * JobID State ExitCode
     * ------------ ----------- --------
     * %job_id% COMPLETED 0:0
     * %job_id%.batch COMPLETED 0:0
     *
     * JobID State ExitCode
     * ------------ ----------- --------
     * %job_id% COMPLETED 0:0
     * %job_id%.batch COMPLETED 0:0
     * %job_id%.0 COMPLETED 0:0
     *
     * Data is extracted from %job_id% and job_id%.batch rows
     *
     * @param array[] $job: output from transform_output_to_named_array()       
     */
    private function parse_job_with_dot_bash($job)
    {

        $data = array();
        $data["MaxDiskRead"] = 0;
        $data["MaxDiskWrite"] = 0;
        $data["CPUTimeRAW"] = 0;
        $data["MaxVMSize"] = 0;
        $data["ReqTRES"] = 0;
        $data["Start"] = INF;

        foreach ($job as $line) {

            if (strpos($line["JobID"], ".batch")) {
                if (trim($line["Start"]) != "") {
                    $data["Start"] = min($data["Start"], $this->make_timestamp($line["Start"]));
                }
                if (trim($line["MaxDiskRead"]) != "") {
                    $data["MaxDiskRead"] += $this->parse_unit_in_KB($line["MaxDiskRead"]);
                }
                if (trim($line["MaxDiskWrite"]) != "") {
                    $data["MaxDiskWrite"] += $this->parse_unit_in_KB($line["MaxDiskWrite"]);
                }
                if (trim($line["MaxVMSize"]) != "") {
                    $data["MaxVMSize"] += $this->parse_unit_in_KB($line["MaxVMSize"]);
                }
                if (trim($line["CPUTimeRAW"]) != "") {
                    $data["CPUTimeRAW"] += $line["CPUTimeRAW"];
                }
            } else if (strpos($line["JobID"], ".")) {
                //Ignore steps when both .batch and .%steps present
            } else {
                if (trim($line["Partition"]) != "") {
                    $data["Partition"] = $line["Partition"];
                }
                if (trim($line["User"]) != "") {
                    $data["User"] = $line["User"];
                }
                if (trim($line["JobID"]) != "") {
                    $data["JobID"] = $line["JobID"];
                }
                if (trim($line["JobName"]) != "") {
                    $data["JobName"] = $line["JobName"];
                }
                if (trim($line["Submit"]) != "") {
                    $data["Submit"] = $this->make_timestamp($line["Submit"]);
                }
                if (trim($line["End"]) != "") {
                    $data["End"] = $this->make_timestamp($line["End"]);
                }
                if (trim($line["ReqTRES"]) != "") {
                    $data["ReqTRES"] += $this->parse_reqTRES($line["ReqTRES"]);
                }
                if (trim($line["NodeList"]) != "") {
                    $data["NodeList"] = $line["NodeList"];
                }
                if (trim($line["NCPUS"]) != "") {
                    $data["NCPUS"] = $line["NCPUS"];
                }
                if (trim($line["ElapsedRaw"]) != "") {
                    $data["ElapsedRaw"] = $line["ElapsedRaw"];
                }
                if (trim($line["State"]) != "") {
                    $data["State"] = $line["State"];
                }
                if (trim($line["ExitCode"]) != "") {
                    $data["ExitCode"] = $line["ExitCode"];
                }
            }
        }

        // If no MaxVM provided, use reserved
        if (trim($data["MaxVMSize"]) == "") {
            $data["MaxVMSize"] = $data["ReqTRES"];
        } else {
            $data["MaxVMSize"] = $this->parse_unit_in_KB($data["MaxVMSize"]);
        }

        unset($job);
        unset($line);
        $this->save($data);
    }

    /**
     * This job prints >1 lines on sacct
     * It has this pattern, probably from a former Slurm version:
     *
     * JobID State ExitCode
     * ------------ ----------- --------
     * %job_id% COMPLETED 0:0
     * %job_id%.0 COMPLETED 0:0
     *
     * As it has no %job_id%.batch, aggregate steps data
     *
     * @param array[] $job: output from transform_output_to_named_array().      
     */
    private function parse_steps_only_job($job)
    {
        
        $data = array();
        $data["MaxDiskRead"] = 0;
        $data["MaxDiskWrite"] = 0;
        $data["CPUTimeRAW"] = 0;
        $data["MaxVMSize"] = 0;
        $data["ReqTRES"] = 0;
        $data["Start"] = INF;

        foreach ($job as $line) {

            if (strpos($line["JobID"], ".")) {
                // Steps rows
                if (trim($line["Start"]) != "") {
                    $data["Start"] = min($data["Start"], $this->make_timestamp($line["Start"]));
                }
                if (trim($line["MaxDiskRead"]) != "") {
                    $data["MaxDiskRead"] += $this->parse_unit_in_KB($line["MaxDiskRead"]);
                }
                if (trim($line["MaxDiskWrite"]) != "") {
                    $data["MaxDiskWrite"] += $this->parse_unit_in_KB($line["MaxDiskWrite"]);
                }
                if (trim($line["MaxVMSize"]) != "") {
                    $data["MaxVMSize"] += $this->parse_unit_in_KB($line["MaxVMSize"]);
                }
                if (trim($line["CPUTimeRAW"]) != "") {
                    $data["CPUTimeRAW"] += $line["CPUTimeRAW"];
                }
            } else {
                // Parent rows
                if (trim($line["Partition"]) != "") {
                    $data["Partition"] = $line["Partition"];
                }
                if (trim($line["User"]) != "") {
                    $data["User"] = $line["User"];
                }
                if (trim($line["JobID"]) != "") {
                    $data["JobID"] = $line["JobID"];
                }
                if (trim($line["JobName"]) != "") {
                    $data["JobName"] = $line["JobName"];
                }
                if (trim($line["Submit"]) != "") {
                    $data["Submit"] = $this->make_timestamp($line["Submit"]);
                }
                if (trim($line["End"]) != "") {
                    $data["End"] = $this->make_timestamp($line["End"]);
                }
                if (trim($line["ReqTRES"]) != "") {
                    $data["ReqTRES"] += $this->parse_reqTRES($line["ReqTRES"]);
                }
                if (trim($line["NodeList"]) != "") {
                    $data["NodeList"] = $line["NodeList"];
                }
                if (trim($line["NCPUS"]) != "") {
                    $data["NCPUS"] = $line["NCPUS"];
                }
                if (trim($line["ElapsedRaw"]) != "") {
                    $data["ElapsedRaw"] = $line["ElapsedRaw"];
                }
                if (trim($line["State"]) != "") {
                    $data["State"] = $line["State"];
                }
                if (trim($line["ExitCode"]) != "") {
                    $data["ExitCode"] = $line["ExitCode"];
                }
            }
        }

        // If no MaxVM provided, use reserved
        if (trim($data["MaxVMSize"]) == "") {
            $data["MaxVMSize"] = $data["ReqTRES"];
        } else {
            $data["MaxVMSize"] = $this->parse_unit_in_KB($data["MaxVMSize"]);
        }

        unset($job);
        unset($line);
        $this->save($data);
    }

    /**
     * Get reserved memory for a job
     *
     * @param string $ReqTRES: sacct output for ReqTRES fields, e.g: billing=3,cpu=3,mem=4000M,node=1
     * @return int number: total memory reserved as mem * core * node
     */
    private function parse_reqTRES(string $ReqTRES): int
    {
        $reqTRESSplit = explode(",", $ReqTRES);
        $mem = 1;
        $node = 1;
        foreach ($reqTRESSplit as $value) {
            if (strpos($value, "mem") !== false) {
                $value_split = explode("=", $value);
                $mem = $this->parse_unit_in_KB($value_split[1]);
            } else if (strpos($value, "node") !== false) {
                $value_split = explode("=", $value);
                $node = $value_split[1];
            }
        }
        return $mem * $node;
    }

    /**
     * Transform memory values to KB
     *
     * @param string $string_in: a number in the form [0-9]+{K,M,G,T,P}{1}
     * @return int number: parsed value for string_in in KB
     */
    private function parse_unit_in_KB($string_in): int
    {
        if (trim($string_in) == "" || $string_in == 0) {
            return 0;
        }

        preg_match("/([0-9]+[\.]?[0-9]+)([A-Z]?)[B]?/", $string_in, $matches);
        if (!isset($matches[2])) {
            if (trim($matches[1]) == "" || $matches[1] == 0) {
                return 0;
            } else {
                return $matches[1];
            }
        }
        switch ($matches[2]) {
            case "K":
                return $matches[1];
                break;
            case "M":
                return $matches[1] * 1024;
                break;
            case "G":
                return $matches[1] * 1048576;
                break;
            case "T":
                return $matches[1] * 1073741824;
                break;
            default:
                if (trim($matches[1]) == "" || $matches[1] == 0) {
                        return 0;
                } else {
                        return $matches[1];
                }
                break;
        }
    }

    /**
     * Transform slurm date to unix timestamp
     *
     * @param string $slurm_data: a date in the form YYYY-MM-DDTHH:mm:ss
     * @return int number: unix timestamp
     */
    private function make_timestamp(string $slurm_date): int
    {
        preg_match('/([0-9]{4})-([0-9]{2})-([0-9]{2})T([0-9]{2}):([0-9]{2}):([0-9]{2})/', $slurm_date, $matches);
        return mktime($matches[4], $matches[5], $matches[6], $matches[2], $matches[3], $matches[1]);
    }

    /**
     * Raw save into database, bypassing MVC framework classes
     *
     * @param array $data: information to save @database
     */
    private function save(array $data)
    {
        if (!isset($data["Partition"]) || trim($data["Partition"]) == "") {
            trigger_error("Discarting job " . $data["JobID"] . " due to missing Partition");
            unset($data);
            return;
        }
        $db = new database();

        $query = "INSERT INTO data (";
        foreach ($this->fields as $field) {
            $query .= "`$field`, ";
        }
        $query = rtrim(rtrim($query), ",");

        $query .= ") VALUES (";
        foreach ($this->fields as $field) {
            $query .= "'" . $data[$field] . "', ";
        }
        $query = rtrim(rtrim($query), ",");

        $query .= ") ON DUPLICATE KEY UPDATE ";

        foreach ($this->fields as $field) {
            $query .= "`$field` = '". $data[$field] . "', ";
        }
        $query = rtrim(rtrim($query), ",");

        if (!$db->query($query)) {
            //die($query);
            trigger_error("Unable to insert data for job " . $data["JobID"] . ". Error:" . $db->error(), E_USER_ERROR); 
        }

        unset($data);
        // echo $query.PHP_EOL;
    }
}
